﻿--[[
Name: LibGuildBankComm-1.0
Revision: $Revision: 14 $
Author(s): Myrroddin of Llane
Description: Saves guild bank contents and sends/receives updates over the addon channel
Dependencies: LibStub, CallbackHandler-1.0, ChatThrottleLib, AceSerializer-3.0, LibCompress
]]

local MAJOR_VERSION = "LibGuildBankComm-1.0"
local MINOR_VERSION = 90000 + tonumber(("$Revision: 14 $"):match("%d+"))

if not LibStub then error(MAJOR_VERSION .. " requires LibStub.") end
local lib, oldversion = LibStub:NewLibrary(MAJOR_VERSION, MINOR_VERSION)
if not lib then return end

local _G = getfenv(0)

local AceSerializer = LibStub("AceSerializer-3.0", true)
local CBH = LibStub("CallbackHandler-1.0")
local AceComm = LibStub("AceComm-3.0")
local ChatThrottleLib = ChatThrottleLib
local LibCompress = LibStub("LibCompress")

lib.eventFrame = lib.eventFrame or CreateFrame("Frame")
lib.eventFrame:SetScript("OnEvent", function(self, event, ...)
	self[event](self, ...)
end)
lib.eventFrame:UnregisterAllEvents()

AceComm:Embed(lib)
lib:RegisterComm("gBankComm", "OnCommReceived")

lib.guildBank = lib.guildBank or {}
local oldRank
local guildRank

lib.callbacks = lib.callbacks or CBH:New(lib)

local recHash -- received hash
function lib:OnCommReceived(prefix, msg, distribution, sender, priority)
	if prefix ~= "gBankComm" then return end
    if sender == "player" then return end
	
	for cmd, recComm in msg:gmatch(("%a+")("%a+")) do
				
		if cmd == "REQUEST" then
			-- serialize lib.guildBank into a string
			-- "zip up" the serialized string
			-- create a hash for comparison
			-- send the hash with higher prio than the serialized string and the serialized string			
			local serialized_data, serialization_error = AceSerializer:Serialize(lib.guildBank)
			if serialization_error then
				return print("Error with serialization: ", serialization_error)
			end
			
			local compressed_data, compression_error = LibCompress:Compress(serialized_data)
			if compression_error then
				return print("Error during compression: ", compression_error)
			end
			
			local hash = LibCompress:fcs32init()
			hash = LibCompress:fcs32update(hash, "UPDATE")
			hash = LibCompress:fcs32update(hash, serialized_data)
			hash = LibCompress:fcs32final(hash)
			
			-- Construct a comm string for the hash.
			local commHash = "HASH " .. hash
			AceComm:SendCommMessage("gBankComm", commHash, "WHISPER", sender, "ALERT")
			
			-- Construct a comm string for the compressed data.
			local commString = "UPDATE " .. compressed_data
			AceComm:SendCommMessage("gBankComm", commString, "WHISPER", sender, "BULK")
			--@debug
			print("Sending bank info to "..tostring(sender))
			--@end-debug			
		elseif cmd == "UPDATE" then
			-- create a new hash
			local compare_hash = LibCompress:fcs32init()
			compare_hash = LibCompress:fcs32upate(compare_hash, "UPDATE")
			compare_hash = LibCompress:fcs32upate(compare_hash, recComm)
			compare_hash = LibCompress:fcs32final(compare_hash)
			compare_hash = compare_hash..sender
			-- compare hashes
			-- if good, deserialize, decompress and check for errors
			-- only update lib.guildBank locally if no errors
			if compare_hash == recHash then
				local decompressed_data, decompression_error = LibCompress:Decompress(recComm)
				if decompression_error then
					return print("Error with decompression: ", decompression_error)
				end
				
				local deserialized_data, deserialization_error = AceSerializer:Deserialize(decompressed_data)			
				if deserialization_error then
					return print("Error during deserialization: ", deserialization_error)
				else				
					lib.guildBank = deserialized_data
					lib:FilterPages()
				end
			else
				return print("Received hash does not match generated hash. Not updating.")
			end
		elseif cmd == "HASH" then
			recHash = "UPDATE "..recComm..sender
		end
	end
end

function lib:FilterPages()
	local guildName = GetGuildInfo("player")
	local pName = UnitName("player")
	_G.GuildControlSetRank(guildRank)
	for page = 1, lib.guildBank.numTabs do
		local canView = GetGuildBankTabPermissions(page)
		if canView ~= 1 then
			lib.guildBank[page].links = nil -- if the player can no longer see a page, nil the info
			lib.guildBank[page].stacks = nil
		end
	end
	-- the player either can see new pages, or can no longer see pages. in any case, need to update addon
	local bankData = lib.guildBank
	lib.callbacks:Fire("GuildBankComm_ContentsChanged", pName, bankData, guildName)
	--[===[@debug@
	print(pName.." from "..tostring(guildName).." has fired GuildBankComm_ContentsChanged because of rank change")
	--@end-debug@]===]	
end

function lib:RequestDataAboutPlayer()
	local _, _, rankIndex = GetGuildInfo("player")
	guildRank = rankIndex
	--[===[@debug@
	print("Player's guild rank is "..guildRank)
	--@end-debug@]===]
	if oldRank ~= guildRank then
		AceComm:SendCommMessage("gBankComm", "REQUEST", "GUILD", nil)
		oldRank = guildRank		
		lib:FilterPages()
	end
end

function lib.eventFrame:GUILD_RANKS_UPDATE(unit)
	if unit ~= "player" then return end
	lib:RequestDataAboutPlayer()
end

function lib.eventFrame:PLAYER_GUILD_UPDATE()
	if not IsInGuild() then -- player left guild			
		local pName = UnitName("player")
		wipe(lib.guildBank)
		local bankData = lib.guildBank
		lib.callbacks:Fire("GuildBankComm_PlayerGuildless", pName, bankData)
	end
end

function lib.eventFrame:GUILDBANKFRAME_OPENED()	
	lib.guildBank.numTabs = GetNumGuildBankTabs()
	lib.guildBank.guildFunds = GetGuildBankMoney()
	--[===[@debug@
	print("Guild Bank frame open")
	--@end-debug@]===]
	
	local name, icon
	for page = 1, lib.guildBank.numTabs do
		name, icon = GetGuildBankTabInfo(page)
		
		if not lib.guildBank[page] then
			lib.guildBank[page] = {
				links = {},  --store the slot item links here
				stacks = {}, --store the slot stack counts here
				name = name,
				texture = icon
			}
		end
	end
end

function lib.eventFrame:GUILDBANK_UPDATE_MONEY()
	lib.guildBank.guildFunds = GetGuildBankMoney()
	--[===[@debug@
	local gName = GetGuildInfo("player")
	print(gName.." has "..lib.guildBank.guildFunds.." in copper")
	--@end-debug@]===]
end

function lib.eventFrame:GUILDBANKFRAME_CLOSED()
	-- Attempt to serialize the data.
	local serialized_data, serialization_error = AceSerializer:Serialize(lib.guildBank)

	-- If serialization failed, spit out a message explaining why, and don't try to continue.
	if serialization_error then
		return print("Error during serialization: ", serialization_error)
	end

	-- Compress the serialized data.
	local compressed_data, compression_error = LibCompress:Compress(serialized_data)
	if compression_error then
		return print ("Error with compression: ", compression_error)
	end

	-- Get a hash for the serialized data.
	local hash = LibCompress:fcs32init()
	hash = LibCompress:fcs32update(hash, "UPDATE")
	hash = LibCompress:fcs32update(hash, serialized_data)
	hash = LibCompress:fcs32final(hash)

	-- Construct a comm string for the hash.
	local commHash = "HASH " .. hash

	-- Do stuff with the hash comm string.
	print("commHash has a length of", commHash:len())

	AceComm:SendCommMessage("gBankComm", commHash, "GUILD", nil, "ALERT")

	-- Construct a comm string for the compressed data.
	local commString = "UPDATE " .. compressed_data
	
	-- Do stuff with the data comm string.
	print("commString has a length of", commString:len())
	AceComm:SendCommMessage("gBankComm", commString, "GUILD", nil, "BULK")
end

function lib.eventFrame:GUILDBANK_UPDATE_TABS()
	lib.guildBank.numTabs = GetNumGuildBankTabs()
end

local num, itemString
function lib.eventFrame:GUILDBANKBAGSLOTS_CHANGED()
	if (GuildBankFrame and GuildBankFrame:IsVisible()) or (BagnonFrameguildbank and BagnonFrameguildbank:IsVisible()) then
		local page
		if not page then
			page = GetCurrentGuildBankTab()
			--[===[@debug@
			print("Looking at page "..tostring(page))
			--@end-debug@]===]
		end
		--98 slots on a page
		for slot = 1, 98 do
			--see if the slot has an item in it
			if GetGuildBankItemLink(page, slot) then
				--if there is link, parse the ID and place that in table w/ key = slot
				_, _, itemString = string.find(GetGuildBankItemLink(page, slot), "^|c%x+|H(.+)|h%[.+%]")
				lib.guildBank[page].links[slot] = itemString
				
				--get item info too, need for stack sizes in image
				_, num = GetGuildBankItemInfo(page, slot)
				lib.guildBank[page].stacks[slot] = num
			else
				-- that slot is empty
				lib.guildBank[page].links[slot] = nil
				lib.guildBank[page].stacks[slot] = nil
			end
			num, itemString = nil, nil
		end
		-- time stamp
		lib.guildBank[page].lastScan = time()
	end
end

function lib.eventFrame:GUILD_ROSTER_UPDATE(unit, forcedUpdate)
	if unit ~= "player" then return end
	if not forcedUpdate then return end -- we only care when GuildRoster() is specifically called during PLAYER_LOGIN
	oldRank = _, _, GetGuildInfo("player")
	lib:RequestDataAboutPlayer()
	--[===[@debug@
	print("Old rank: "..tostring(oldRank))
	--@end-debug@]===]
end

function lib.eventFrame:PLAYER_LOGIN()	
	if IsInGuild() then
		_G.GuildRoster()
	end
end

function lib:Start()
	lib.eventFrame:RegisterEvent("PLAYER_GUILD_UPDATE")
	lib.eventFrame:RegisterEvent("GUILD_RANKS_UPDATE")
	lib.eventFrame:RegisterEvent("GUILDBANKBAGSLOTS_CHANGED")
	lib.eventFrame:RegisterEvent("GUILDBANKFRAME_OPENED")
	lib.eventFrame:RegisterEvent("GUILDBANKFRAME_CLOSED")
	lib.eventFrame:RegisterEvent("GUILDBANK_UPDATE_TABS")
	lib.eventFrame:RegisterEvent("GUILDBANK_UPDATE_MONEY")	
	lib.eventFrame:RegisterEvent("PLAYER_LOGIN")
	lib.eventFrame:RegisterEvent("GUILD_ROSTER_UPDATE")
end
lib:Start()